// Copyright JGX Software Solutions, 2018.

// THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
// THIS SOFTWARE.

#include <vector>
#include "Entities.h"

using namespace std;

void GenerateEdgeElementChain(const vector<node *> &, const vector<node *> &, vector<vector<node *>> &);
void GenerateNodeChain(int d, double[2], double[2], vector<node *> &, vector<node *> &);
void GenerateNodeChain(int d, double[2], double[2], node *, vector<node *> &, vector<node *> &);
void GenerateNodeChain(node *, int d, double[2], double[2], vector<node *> &, vector<node *> &);
void GenerateNodeChain(node *, const vector<node *> &, vector<node *> &);
void GenerateNodeChain(const vector<node *> &, node *, vector<node *> &);
void GenerateNodeChain(node *, const vector<node *> &, node *, vector<node *> &);


void GenerateMappedMesh4Sides(int d1, int d2, int d3, int d4,
	double size[2], double anchor[2],
	vector<node *> & uu1,
	vector<node *> & uu2,
	vector<node *> & vv1,
	vector<node *> & vv2,
	vector<vector<node *>> & ee,
	vector<node *> & nn
) {

	if (d1 <= 0 || d2 <= 0 || d3 <= 0 || d4 <= 0) return;

	if (d1 == d3 && d2 == d4) {
		// This is just a case of generating a d1 x d2 grid.

		vector<node *> nlist(4);
		element * pElement;

		double stepx = size[0] / (double)d1;
		double stepy = size[1] / (double)d2;

		double x, y;

		node * pNode;

		// Generate some nodes.

		vector<node *> newnodes((d1 + 1)*(d2 + 1));

		int iptr = 0;

		x = anchor[0];
		for (int i = 0; i <= d1; ++i) {
			y = anchor[1];
			for (int j = 0; j <= d2; ++j) {
				pNode = newNode();
				pNode->putCoords({ x, y, 0.0 });
				y += stepy;
				newnodes[iptr] = pNode;
				nn.push_back(pNode);
				iptr++;
			}
			x += stepx;
		}

		int DY = d2 + 1;

		for (int i = 0; i < d1; ++i) {
			y = anchor[1];
			for (int j = 0; j < d2; ++j) {

				nlist[0] = newnodes[i*DY + j];
				nlist[1] = newnodes[(i + 1)*DY + j];
				nlist[2] = newnodes[(i + 1)*DY + j + 1];
				nlist[3] = newnodes[i*DY + j + 1];

				ee.push_back(nlist);

				pElement = newElement();
				pElement->putNodes(nlist);
			}
		}

		// Initialize the boundary node vectors.

		uu1.resize(d1 + 1);
		uu2.resize(d1 + 1);
		vv1.resize(d2 + 1);
		vv2.resize(d2 + 1);

		for (int i = 0; i <= d1; ++i) {
			uu1[i] = newnodes[i*DY];
			uu2[i] = newnodes[i*DY + d2];
		}

		for (int i = 0; i <= d2; ++i) {
			vv1[i] = newnodes[i];
			vv2[i] = newnodes[d1*DY + i];
		}

	}
	else {

		int cornerCount[4];
		cornerCount[0] = d4 + d1;
		cornerCount[1] = d1 + d2;
		cornerCount[2] = d2 + d3;
		cornerCount[3] = d3 + d4;

		// If there are two minimal (equal) corner counts, then the second
		// corner is given by imin2 >= 0.

		int imin = 0;
		int imin2 = -1;
		int iccmin = cornerCount[0];
		for (int i = 1; i < 4; ++i) {
			if (cornerCount[i] < iccmin) {
				iccmin = cornerCount[i];
				imin = i;
				imin2 = -1;
			}
			else if (cornerCount[i] == iccmin) {
				imin2 = i;
			}
		}

		if (imin2 == 3 && imin == 0) {
			imin = 3;
			imin2 = 0;
		}

		double stepx, stepy;
		double anchor_new[2], size_new[2];

		double anch1[] = { anchor[0], anchor[1] + size[1] };
		double anch2[] = { anchor[0] + size[0], anchor[1] };

		if (imin2 >= 0) {
			// Double minimal corner count.
			switch (imin) {
			case 0:
				//
				stepx = size[0] / (double)d3;
				stepy = size[1] / (double)d2;
				d3 -= 2;
				anchor_new[0] = anchor[0] + stepx;
				anchor_new[1] = anchor[1] + stepy;
				size_new[0] = size[0] - 2.0*stepx;
				size_new[1] = size[1] - stepy;
				break;
			case 1:
				stepx = size[0] / (double)d1;
				stepy = size[1] / (double)d4;
				d4 -= 2;
				anchor_new[0] = anchor[0];
				anchor_new[1] = anchor[1] + stepy;
				size_new[0] = size[0] - stepx;
				size_new[1] = size[1] - 2.0*stepy;
				break;
			case 2:
				stepx = size[0] / (double)d1;
				stepy = size[1] / (double)d2;
				d1 -= 2;
				anchor_new[0] = anchor[0] + stepx;
				anchor_new[1] = anchor[1];
				size_new[0] = size[0] - 2.0*stepx;
				size_new[1] = size[1] - stepy;
				break;
			case 3:
				stepx = size[0] / (double)d1;
				stepy = size[1] / (double)d2;
				d2 -= 2;
				anchor_new[0] = anchor[0] + stepx;
				anchor_new[1] = anchor[1] + stepy;
				size_new[0] = size[0] - stepx;
				size_new[1] = size[1] - 2.0*stepy;
				break;
			}

			vector<node *> uun1;
			vector<node *> uun2;
			vector<node *> vvn1;
			vector<node *> vvn2;

			GenerateMappedMesh4Sides(d1, d2, d3, d4, size_new, anchor_new, uun1, uun2, vvn1, vvn2, ee, nn);

			double dirx[2];
			double diry[2];

			switch (imin) {
			case 0:
				// Note that in this situation d2==d4.
				dirx[0] = size[0] / (double)d1; dirx[1] = 0.0;
				diry[0] = 0.0;  diry[1] = size[1] / (double)d2;

				GenerateNodeChain(d1, anchor, dirx, uu1, nn);
				GenerateNodeChain(uu1[0], d2, anchor, diry, vv1, nn);
				GenerateNodeChain(uu1[d1], d2, anch2, diry, vv2, nn);

				GenerateEdgeElementChain(uu1, uun1, ee);
				GenerateEdgeElementChain(vvn1, vv1, ee);
				GenerateEdgeElementChain(vv2, vvn2, ee);

				GenerateNodeChain(vv1[d2], uun2, vv2[d2], uu2);

				break;
			case 1:
				// Note that in this situation d1==d3.
				dirx[0] = size[0] / (double)d1; dirx[1] = 0.0;
				diry[0] = 0.0;  diry[1] = size[1] / (double)d2;

				GenerateNodeChain(d1, anchor, dirx, uu1, nn);
				GenerateNodeChain(uu1[d1], d2, anch2, diry, vv2, nn);
				GenerateNodeChain(d1, anch1, dirx, vv2[d2], uu2, nn);

				GenerateEdgeElementChain(uu1, uun1, ee);
				GenerateEdgeElementChain(vv2, vvn2, ee);
				GenerateEdgeElementChain(uun2, uu2, ee);

				GenerateNodeChain(uu1[0], vvn1, uu2[0], vv1);

				break;
			case 2:
				// Note that in this situation d2==d4.
				dirx[0] = size[0] / (double)d3; dirx[1] = 0.0;
				diry[0] = 0.0; diry[1] = size[1] / (double)d2;

				GenerateNodeChain(d3, anch1, dirx, uu2, nn);
				GenerateNodeChain(d2, anch2, diry, uu2[d3], vv2, nn);
				GenerateNodeChain(d2, anchor, diry, uu2[0], vv1, nn);

				GenerateEdgeElementChain(vvn1, vv1, ee);
				GenerateEdgeElementChain(uun2, uu2, ee);
				GenerateEdgeElementChain(vv2, vvn2, ee);

				GenerateNodeChain(vv1[0], uun1, vv2[0], uu1);

				break;
			case 3:
				// Note that in this situation d1==d3.
				dirx[0] = size[0] / (double)d3; dirx[1] = 0.0;
				diry[0] = 0.0; diry[1] = size[1] / (double)d4;

				GenerateNodeChain(d1, anchor, dirx, uu1, nn);
				GenerateNodeChain(uu1[0], d4, anchor, diry, vv1, nn);
				GenerateNodeChain(vv1[d4], d1, anch1, dirx, uu2, nn);

				GenerateEdgeElementChain(uu1, uun1, ee);
				GenerateEdgeElementChain(uun2, uu2, ee);
				GenerateEdgeElementChain(vvn1, vv1, ee);

				GenerateNodeChain(uu1[d1], vvn2, uu2[d1], vv2);

				break;
			}

		}
		else {
			// Single minimal corner count.
			switch (imin) {
			case 0:
				//
				stepx = size[0] / (double)d3;
				stepy = size[1] / (double)d2;
				d3--;
				d2--;
				anchor_new[0] = anchor[0] + stepx;
				anchor_new[1] = anchor[1] + stepy;
				break;
			case 1:
				stepx = size[0] / (double)d3;
				stepy = size[1] / (double)d4;
				d3--;
				d4--;
				anchor_new[0] = anchor[0];
				anchor_new[1] = anchor[1] + stepy;
				break;
			case 2:
				stepx = size[0] / (double)d1;
				stepy = size[1] / (double)d4;
				d1--;
				d4--;
				anchor_new[0] = anchor[0];
				anchor_new[1] = anchor[1];
				break;
			case 3:
				stepx = size[0] / (double)d1;
				stepy = size[1] / (double)d2;
				d1--;
				d2--;
				anchor_new[0] = anchor[0] + stepx;
				anchor_new[1] = anchor[1];
				break;
			}

			size_new[0] = size[0] - stepx;
			size_new[1] = size[1] - stepy;

			vector<node *> uun1;
			vector<node *> uun2;
			vector<node *> vvn1;
			vector<node *> vvn2;

			GenerateMappedMesh4Sides(d1, d2, d3, d4, size_new, anchor_new, uun1, uun2, vvn1, vvn2, ee, nn);

			double dirx[2];
			double diry[2];

			switch (imin) {
			case 0:
				dirx[0] = size[0] / (double)d1; dirx[1] = 0.0;
				diry[0] = 0.0; diry[1] = size[1] / (double)d4;

				GenerateNodeChain(d1, anchor, dirx, uu1, nn);
				GenerateNodeChain(uu1[0], d4, anchor, diry, vv1, nn);
				GenerateEdgeElementChain(uu1, uun1, ee);
				GenerateEdgeElementChain(vvn1, vv1, ee);
				GenerateNodeChain(uu1[d1], vvn2, vv2);
				GenerateNodeChain(vv1[d4], uun2, uu2);

				break;
			case 1:
				dirx[0] = size[0] / (double)d1; dirx[1] = 0.0;
				diry[0] = 0.0; diry[1] = size[1] / (double)d2;

				GenerateNodeChain(d1, anchor, dirx, uu1, nn);
				GenerateNodeChain(uu1[d1], d2, anch2, diry, vv2, nn);
				GenerateEdgeElementChain(uu1, uun1, ee);
				GenerateEdgeElementChain(vv2, vvn2, ee);
				GenerateNodeChain(uun2, vv2[d2], uu2);
				GenerateNodeChain(uu1[0], vvn1, vv1);

				break;
			case 2:
				dirx[0] = size[0] / (double)d3; dirx[1] = 0.0;
				diry[0] = 0.0; diry[1] = size[1] / (double)d2;

				GenerateNodeChain(d3, anch1, dirx, uu2, nn);
				GenerateNodeChain(d2, anch2, diry, uu2[d3], vv2, nn);
				GenerateEdgeElementChain(uun2, uu2, ee);
				GenerateEdgeElementChain(vv2, vvn2, ee);
				GenerateNodeChain(uun1, vv2[0], uu1);
				GenerateNodeChain(vvn1, uu2[0], vv1);

				break;
			case 3:
				dirx[0] = size[0] / (double)d3; dirx[1] = 0.0;
				diry[0] = 0.0; diry[1] = size[1] / (double)d4;

				GenerateNodeChain(d3, anch1, dirx, uu2, nn);
				GenerateNodeChain(d4, anchor, diry, uu2[0], vv1, nn);
				GenerateEdgeElementChain(uun2, uu2, ee);
				GenerateEdgeElementChain(vvn1, vv1, ee);
				GenerateNodeChain(vv1[0], uun1, uu1);
				GenerateNodeChain(vvn2, uu2[d3], vv2);

				break;
			}

		}

	}

}


void GenerateNodeChain(int d, double anchor[2], double step[2], vector<node *> & uu, vector<node *> & nn) {

	uu.resize(d + 1);

	node * pNode;
	double x, y;

	x = anchor[0];
	y = anchor[1];

	for (int i = 0; i <= d; ++i) {
		pNode = newNode();
		nn.push_back(pNode);
		pNode->putCoords({ x, y, 0.0 });
		uu[i] = pNode;
		x += step[0];
		y += step[1];
	}

}

void GenerateNodeChain(int d, double anchor[2], double step[2], node * pEnd, vector<node *> & uu, vector<node *> & nn) {

	uu.resize(d + 1);

	node * pNode;
	double x, y;

	x = anchor[0];
	y = anchor[1];

	for (int i = 0; i < d; ++i) {
		pNode = newNode();
		nn.push_back(pNode);
		pNode->putCoords({ x, y, 0.0 });
		uu[i] = pNode;
		x += step[0];
		y += step[1];
	}

	uu[d] = pEnd;

}

void GenerateNodeChain(node * pStart, int d, double anchor[2], double step[2], vector<node *> & uu, vector<node *> & nn) {

	uu.resize(d + 1);

	uu[0] = pStart;

	node * pNode;
	double x, y;

	x = anchor[0] + step[0];
	y = anchor[1] + step[1];

	for (int i = 0; i < d; ++i) {
		pNode = newNode();
		nn.push_back(pNode);
		pNode->putCoords({ x, y, 0.0 });
		uu[i + 1] = pNode;
		x += step[0];
		y += step[1];
	}

}

void GenerateNodeChain(node * pStart, const vector<node *> & uun, vector<node *> & uu) {

	uu.resize(uun.size() + 1);

	uu[0] = pStart;

	for (size_t i = 0; i < uun.size(); ++i) {
		uu[i + 1] = uun[i];
	}

}

void GenerateNodeChain(const vector<node *> & uun, node * pEnd, vector<node *> & uu) {

	uu.resize(uun.size() + 1);

	for (size_t i = 0; i < uun.size(); ++i) {
		uu[i] = uun[i];
	}

	uu[uun.size()] = pEnd;

}

void GenerateNodeChain(node * pStart, const vector<node *> & uun, node * pEnd, vector<node *> & uu) {

	uu.resize(uun.size() + 2);

	uu[0] = pStart;

	for (unsigned i = 0; i < uun.size(); ++i) {
		uu[i + 1] = uun[i];
	}

	uu[uun.size() + 1] = pEnd;

}





void GenerateEdgeElementChain(const vector<node *> & uu, const vector<node *> & uun, vector<vector<node *>> & ee) {

	vector<node *> nlist(4);
	element * pElement;

	for (size_t i = 0; i < uu.size() - 1; ++i) {

		nlist[0] = uu[i];
		nlist[1] = uu[i + 1];
		nlist[2] = uun[i + 1];
		nlist[3] = uun[i];

		if (nlist[0]->getName() == "19") {
			int y = 9;
		}

		pElement = newElement();
		pElement->putNodes(nlist);

		ee.push_back(nlist);
	}

}


