// Copyright JGX Software Solutions, 2018.

// THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
// THIS SOFTWARE.

#include <vector>
#include <iterator>
#include <fstream>

#include "Entities.h"
#include "V3.h"

using namespace std;

void quadMesh(int[4]);
void exportSurfaceMesh(ofstream &);

int main()
{
	int isegs[4] = { 3, 5, 6, 4 };
	quadMesh(isegs);

	ofstream fs("mesh.jdg");
	exportSurfaceMesh(fs);
	fs.close();

	terminateEntities();

    return 0;
}


void SmoothMesh(V3, const vector<vector<node *>> &, const vector<node *> &);

void CalculateNodalConnectivities(const vector<vector<node *>> &);

void GenerateMappedMesh4Sides(int, int, int, int, double[2], double[2],
	vector<node *> &, vector<node *> &, vector<node *> &, vector<node *> &,
	vector<vector<node *>> &, vector<node *> &
);


void quadMesh(int isegs[4]) {

	clearMesh();

	// Generate the mesh on a unit square anchored at the origin.

	double anchor[2] = { 0.0, 0.0 };
	double size[2] = { 1.0, 1.0 };

	vector<node *> uu1(isegs[0] + 1);
	vector<node *> uu2(isegs[2] + 1);
	vector<node *> vv1(isegs[3] + 1);
	vector<node *> vv2(isegs[1] + 1);

	vector<vector<node *>> elements;
	vector<node *> nodes;

	GenerateMappedMesh4Sides(isegs[0], isegs[1], isegs[2], isegs[3], size, anchor, uu1, uu2, vv1, vv2, elements, nodes);

	for (auto p = uu1.begin(); p < uu1.end(); ++p) (*p)->setFixed();
	for (auto p = uu2.begin(); p < uu2.end(); ++p) (*p)->setFixed();
	for (auto p = vv1.begin(); p < vv1.end(); ++p) (*p)->setFixed();
	for (auto p = vv2.begin(); p < vv2.end(); ++p) (*p)->setFixed();

	CalculateNodalConnectivities(elements);

	SmoothMesh(V3(0.0, 0.0, 1.0), elements, nodes);
}


#include "V3.h"
#include "Entities.h"
#include <vector>
#include <algorithm>

using namespace std;

V3 minangAdjustment(V3, V3, V3, V3);
V3 aspectRatioAdjustment(V3, V3, V3, V3);

void SmoothMesh(V3 normal, const vector<vector<node *>> & ee, const vector<node *> & nn) {

	// normal  the normal aabout which the elements are orientated.
	// ee      a constant list of elements.
	// nn      a constant list of nodes.

	// For any node that is not fixed (i.e. not on the boundary), initialize its
	// smoothing vector, which is the positional adjustment that has been 
	// calculated to improve the overall quality of the referencing elements.

	for (auto q = ee.begin(); q < ee.end(); ++q) {
		auto rprev = q->end() - 1;
		for (auto r = q->begin(); r < q->end(); ++r) {
			if ((*r)->isFixed()) continue;
			auto rnext = r + 1;
			if (r == q->end() - 1) rnext = q->begin();

			(*r)->putSmoothVector({ 0.0, 0.0, 0.0 });
			(*r)->setAdjusted();

			rprev = r;
		}
	}


	// Iteratively adjust the position of non-fixed nodes in order to
	// improve the quality of the referencing elements. This will 
	// continue for a maximum of 30 iterations or until the
	// maxiumum adjustment to any fixed node falls below 0.0001.

	int ite = 0;

	double maxAdjust = 0.0001;

	while (ite < 30 && maxAdjust >= 0.0001) {
		ite++;

		for (auto q = ee.begin(); q < ee.end(); ++q) {

			auto rprev = q->end() - 1;
			for (auto r = q->begin(); r < q->end(); ++r) {
				if ((*r)->isFixed()) {
					rprev = r;
					continue;
				}

				auto rnext = r + 1;
				if (r == q->end() - 1) rnext = q->begin();

				// Calculate the nodal adjustment needed to improve the quality of the element by itself.

				V3 adjustVector = minangAdjustment((*rprev)->getCoords(), (*r)->getCoords(), (*rnext)->getCoords(), normal);

				// The node is shared by other elements, so the contribution of this adjustment to the overall
				// adjustment is pro-rata.

				adjustVector = adjustVector / (double)((*r)->getConnectivity());

				// Mark the node as requiring an adjustment.

				(*r)->incrementSmoothVector(adjustVector);
				(*r)->putAdjusted(false);

				rprev = r;
			}
		}

		// Nodal adjustment and calculation of the maximum adjustment over the entire
		// mesh.

		double maxAdjust = 0.0;

		for (auto q = ee.begin(); q < ee.end(); ++q) {

			// For each unadjusted node in the element's node list,
			// add the contents of its smoothing vector to its position,
			// if necessary update the maximum adjustment for the entire mesh
			// and thereafter reset its smoothing vector to zero.

			for (auto r = q->begin(); r < q->end(); ++r) {
				if ((*r)->isFixed()) continue;
				if ((*r)->isAdjusted()) continue;

				(*r)->setAdjusted();

				V3 adjustVector = (*r)->getSmoothVector();
				maxAdjust = max(maxAdjust, adjustVector.length());

				adjustVector = adjustVector + (*r)->getCoords();
				(*r)->putCoords(adjustVector);

				(*r)->putSmoothVector({ 0.0, 0.0, 0.0 });
			}
		}
	}
}



V3 minangAdjustment(V3 cprev, V3 ccurr, V3 cnext, V3 normal) {

	V3 p = cnext - cprev;

	V3 res = cprev + cnext + (p ^ normal);
	res = res / 2.0;
	res = res - ccurr;

	return res;

}

V3 aspectRatioAdjustment(V3 cprev, V3 ccurr, V3 cnext, V3 normal) {

	V3 p = cnext - cprev;

	V3 res = cprev + cnext + sqrt(3.0) * (p ^ normal);
	res = res / 2.0;
	res = res - ccurr;

	return res;

}


void CalculateNodalConnectivities(const vector<vector<node *>> & ee) {
	// ee     a constant list of elements.

	// Initialize the connectivity of each referenced node.

	for (auto q = ee.begin(); q < ee.end(); ++q) {

		for (auto r = q->begin(); r < q->end(); ++r) {
			(*r)->putConnectivity(0);
		}
	}

	// For each element increment the connectivity of each node in its list.

	for (auto q = ee.begin(); q < ee.end(); ++q) {

		for (auto r = q->begin(); r < q->end(); ++r) {
			(*r)->incrementConnectivity();
		}
	}

}




